from __future__ import annotations

from fastapi import FastAPI

{{imports}}

app = FastAPI()


{% for operation in operations %}
@app.{{operation.type}}('{{operation.snake_case_path}}', response_model={{operation.response}})
def {{operation.function_name}}({{operation.snake_case_arguments}}) -> {{operation.response}}:
    pass
{% endfor %}